import abc
from abc import abstractmethod
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional

from pydantic import BaseModel, ConfigDict, Field, field_validator, model_serializer

from trano.reporting.utils import to_html_construction, to_html_space, to_html_system

Topic = Literal["Spaces", "Construction", "Systems", "Base"]


class BaseReporting(BaseModel):
    model_config = ConfigDict(populate_by_name=True, extra="forbid")


class BaseTable(abc.ABC, BaseReporting):
    @abstractmethod
    def to_html(self) -> str: ...


class BaseNestedTable(BaseReporting):
    name: Optional[str] = None
    parameters: Dict[str, Any] = Field(default_factory=dict)

    @model_serializer
    def serializer(self) -> Dict[str, Any]:
        if not self.parameters:
            return {}
        return self.parameters | {"name": self.name}


class ContentDocumentation(BaseModel):
    title: Optional[str] = None
    introduction: Optional[str] = None
    conclusions: Optional[str] = None


class BoundaryTable(BaseReporting):
    name: str
    surface: float
    tilt: str
    construction: str
    azimuth: Optional[float] = None

    @field_validator("construction", mode="before")
    @classmethod
    def _construction_validator(cls, value: Dict[str, Any]) -> Optional[str]:
        return value.get("name")


class EmissionTable(BaseNestedTable): ...


class OccupancyTable(BaseNestedTable): ...


class SystemTable(BaseTable, BaseNestedTable):
    def to_html(self) -> str:
        return to_html_system(self.model_dump(exclude_none=True))


class SpaceTable(BaseTable):
    name: str
    parameters: Dict[str, Any]
    external_boundaries: List[BoundaryTable]
    internal_elements: List[BoundaryTable] = Field(default_factory=list)
    emissions: List[EmissionTable] = Field(default_factory=list)
    occupancy: Optional[OccupancyTable] = Field(default_factory=OccupancyTable)

    def to_html(self) -> str:
        return to_html_space(self.model_dump(exclude_none=True))


class LayerTable(BaseReporting):
    name: str
    k: float
    c: float
    rho: float
    epsLw: float  # noqa: N815
    epsSw: float  # noqa: N815
    thickness: float
    solar_transmittance: List[float] = Field(default_factory=list)
    solar_reflectance_outside_facing: List[float] = Field(default_factory=list)
    solar_reflectance_room_facing: List[float] = Field(default_factory=list)
    infrared_transmissivity: Optional[float] = None
    infrared_absorptivity_outside_facing: Optional[float] = None
    infrared_absorptivity_room_facing: Optional[float] = None


class ConstructionTable(BaseTable):
    name: str
    layers: List[LayerTable]
    u_value_frame: Optional[float] = None

    @field_validator("layers", mode="before")
    @classmethod
    def layers_validator(cls, layers: List[Dict[str, Any]]) -> List[LayerTable]:
        return [
            LayerTable(**(layer["material"] | {"thickness": layer["thickness"]}))
            for layer in layers
        ]

    def to_html(self) -> str:
        return to_html_construction(self.model_dump())


class ResultFile(BaseModel):
    path: Path
    type: Literal["openmodelica", "dymola"] = "openmodelica"



from pathlib import Path

import yaml


def create_final_schema(
    parameters_path: Path, trano_final_path: Path, trano_path: Path
) -> None:
    trano = yaml.safe_load(trano_path.read_text())
    parameters = yaml.safe_load(parameters_path.read_text())
    for name, parameter in parameters.items():
        parameter.pop("classes")
        parameter__ = {}
        for k, v in parameter["attributes"].items():
            if "func" not in v:
                v.pop("alias", None)
                parameter__[k] = v
        parameter["attributes"] = parameter__
        trano["classes"][name] = parameter
    yaml.dump(trano, trano_final_path.open("w"))



from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional

import yaml
from jinja2 import Environment, FileSystemLoader
from pydantic import BaseModel, Field

from trano.elements.controller_bus import ControllerBus
from trano.elements.types import DynamicTemplateCategories

if TYPE_CHECKING:
    from trano.elements.base import BaseElement


def _load_components() -> Dict[str, Any]:
    libraries_path = Path(__file__).parent.joinpath("models")
    data: Dict[str, Any] = {"components": []}
    for file in libraries_path.glob("*.yaml"):
        data["components"] += yaml.safe_load(file.read_text()).get("components", [])
    return data


COMPONENTS = _load_components()


class DynamicComponentTemplate(BaseModel):

    template: str
    category: Optional[DynamicTemplateCategories] = None
    function: Callable[[Any], Any] = Field(default=lambda _: {})
    bus: ControllerBus

    def render(
        self, package_name: str, element: "BaseElement", parameters: Dict[str, Any]
    ) -> str:
        ports = list(self.bus.bus_ports(element))
        environment = Environment(
            trim_blocks=True,
            lstrip_blocks=True,
            loader=FileSystemLoader(
                str(Path(__file__).parents[1].joinpath("templates"))
            ),
            autoescape=True,
        )
        environment.filters["enumerate"] = enumerate
        rtemplate = environment.from_string(
            "{% import 'macros.jinja2' as macros %}" + self.template
        )
        component = rtemplate.render(
            element=element,
            package_name=package_name,
            bus_template=self.bus.template,
            bus_ports="\n".join(ports),
            parameters=parameters,
            **self.function(element),
        )

        return component



import platform
import shutil
import subprocess
import tempfile
from contextlib import contextmanager
from pathlib import Path
from typing import Generator, Optional

import docker  # type: ignore
from jinja2 import Environment
from pydantic import BaseModel, Field

from trano.exceptions import DockerNotInstalledError, DockerClientError
from trano.topology import Network


def check_docker_installed() -> None:
    try:
        subprocess.run(
            ["docker", "--version"],  # noqa: S603, S607
            check=True,
        )
    except (subprocess.CalledProcessError, FileNotFoundError) as e:
        raise DockerNotInstalledError(
            "Docker is not installed on the system. Simulation cannot be run."
        ) from e


def client() -> docker.DockerClient:
    check_docker_installed()
    system = platform.system()

    if system in ("Linux", "Darwin"):
        base_url = "unix:///var/run/docker.sock"
    elif system == "Windows":
        base_url = "npipe:////./pipe/docker_engine"
    else:
        raise NotImplementedError(f"Unsupported platform: {system}")
    try:
        client = docker.DockerClient(base_url=base_url)
    except Exception as e:
        raise DockerClientError(
            f"Docker client cannot be initialized with base url: '{base_url}'. "
            f"Simulation cannot be run on your {system} system."
        ) from e
    return client


class SimulationOptions(BaseModel):
    start_time: int = Field(default=0)
    end_time: int = Field(default=2 * 3600 * 24 * 7)
    check_only: bool = Field(default=False)
    tolerance: float = Field(default=1e-4)


class SimulationLibraryOptions(SimulationOptions):
    library_name: str = Field(default="Buildings")


def simulate(
    project_path: Path,
    model_network: Network,
    options: Optional[SimulationOptions] = None,
) -> docker.models.containers.ExecResult:
    client_ = client()
    options = options or SimulationOptions()
    with container(client_, project_path) as container_, create_mos_file(
        model_network, options, project_path
    ) as mos_file_name:
        results = container_.exec_run(cmd=f"omc /simulation/{mos_file_name}")
    return results


def stop_container(client: docker.DockerClient, container_name: str) -> None:
    try:
        container = client.containers.get(container_name)
        if container.attrs["State"]["Status"] == "running":
            container.stop()
        container.remove()
    except docker.errors.NotFound:
        pass


@contextmanager
def container(
    client: docker.DockerClient, project_path: Path
) -> Generator[docker.models.containers.Container, None, None]:
    container_name = "openmodelica"
    stop_container(client, container_name)
    container = client.containers.run(
        "openmodelica/openmodelica:v1.22.4-ompython",
        command="tail -f /dev/null",
        volumes=[
            f"{project_path}:/simulation",
            f"{project_path}/results:/results",
        ],
        detach=True,
        name=container_name,
    )
    configuration_path = Path(__file__).parent / "configure.mos"
    shutil.copy(configuration_path, project_path / "configure.mos")
    container.exec_run(cmd="chmod -R 777 /results")
    container.exec_run(cmd="chmod -R 777 /simulation")
    container.exec_run(cmd="omc /simulation/configure.mos")
    yield container
    container.exec_run(
        cmd='find / -name "*_res.mat" -exec cp {} /results \;'  # noqa: W605
    )
    container.stop()
    container.remove()


@contextmanager
def create_mos_file(
    network: Network, options: SimulationOptions, project_path: Path
) -> Generator[str, None, None]:
    # TODO: do we want this here?
    network.set_weather_path_to_container_path(project_path)
    model = network.model()
    with tempfile.NamedTemporaryFile(
        mode="w", dir=project_path, suffix=".mo"
    ) as temp_model_file, tempfile.NamedTemporaryFile(
        mode="w", dir=project_path, suffix=".mos"
    ) as temp_mos_file:
        Path(temp_model_file.name).write_text(model)
        environment = Environment(autoescape=True)
        if options.check_only:
            template = environment.from_string(
                """
    getVersion();
    loadFile("/simulation/{{model_file}}");
    checkModel({{model_name}}.building);
    """
            )
        else:
            template = environment.from_string(
                f"""
    getVersion();
    loadFile("/simulation/{{{{model_file}}}}");
    checkModel({{{{model_name}}}}.building);
    simulate({{{{model_name}}}}.building,startTime = {options.start_time},
    stopTime = {options.end_time},
    tolerance = {options.tolerance});
    """
            )
        mos_file = template.render(
            model_file=Path(temp_model_file.name).name, model_name=network.name
        )
        Path(temp_mos_file.name).write_text(mos_file)
        yield Path(temp_mos_file.name).name



from typing import TYPE_CHECKING, Any, Dict, List, Type, Union

from buildingspy.io.outputfile import Reader  # type: ignore
from jinja2 import Template

from trano.plot.plot import plot_element, plot_plot_ly

if TYPE_CHECKING:
    from trano.elements import BaseElement, BaseSimpleWall, BaseWall, Space, System
    from trano.reporting.reporting import ModelDocumentation


def to_html_space(data: Dict[str, Any]) -> str:
    template = Template(
        """
    <!DOCTYPE html>
    <html lang="en">
    <head>
        <meta charset="UTF-8">
        <meta name="viewport" content="width=device-width, initial-scale=1.0">
        <title>External Boundaries Table</title>
        <style>
            body {
                font-family: Arial, sans-serif;
            }
            .fancy-table {
                width: 100%;
                border-collapse: collapse;
                margin: 20px 0;
                font-size: 1em;
                text-align: left;
            }
            .fancy-table th, .fancy-table td {
                padding: 12px;
                border: 1px solid #ddd;
            }
            .fancy-table th {
                background-color: #f4f4f4;
                font-weight: bold;
            }
            .fancy-table tr:nth-child(even) {
                background-color: #f9f9f9;
            }
            .fancy-table caption {
                font-size: 1.5em;
                margin-bottom: 10px;
                font-weight: bold;
            }
        </style>
    </head>
    <body>
    {% for parameter_ in [data.parameters,data.occupancy] %}
    <table>
        <thead>
            <tr>
                {% for key in parameter_.keys() %}
                    <th><strong>{{ key }}</strong></th>
                {% endfor %}
            </tr>
        </thead>
        <tbody>
            <tr>
                {% for value in parameter_.values() %}
                    <td>{{ value }}</td>
                {% endfor %}
            </tr>
        </tbody>
    </table>
    {% endfor %}
    <table class="fancy-table">
        <thead>
            <tr>
                <th>Name</th>
                <th>Azimuth</th>
                <th>Construction Name</th>
                <th>Surface</th>
                <th>Tilt</th>
            </tr>
        </thead>
        <tbody>
        {% for boundary in data.external_boundaries + data.internal_elements %}
            <tr>
                <td>{{ boundary.name }}</td>
                <td>{{ boundary.azimuth }}</td>
                <td>{{ boundary.construction }}</td>
                <td>{{ boundary.surface }}</td>
                <td>{{ boundary.tilt }}</td>
            </tr>
        {% endfor %}
        </tbody>

        {% for emission in data.emissions %}
    <table>
        <thead>
            <tr>
                {% for key in emission.keys() %}
                    <th><strong>{{ key }}</strong></th>
                {% endfor %}
            </tr>
        </thead>
        <tbody>
            <tr>
                {% for value in emission.values() %}
                    <td>{{ value }}</td>
                {% endfor %}
            </tr>
        </tbody>
    </table>
    {% endfor %}

    </body>
    </html>
    """
    )

    return template.render(data=data)


def to_html_system(data: Dict[str, Any]) -> str:
    template = Template(
        """
    <!DOCTYPE html>
    <html lang="en">
    <head>
        <meta charset="UTF-8">
        <meta name="viewport" content="width=device-width, initial-scale=1.0">
        <title>External Boundaries Table</title>
        <style>
            body {
                font-family: Arial, sans-serif;
            }
            .fancy-table {
                width: 100%;
                border-collapse: collapse;
                margin: 20px 0;
                font-size: 1em;
                text-align: left;
            }
            .fancy-table th, .fancy-table td {
                padding: 12px;
                border: 1px solid #ddd;
            }
            .fancy-table th {
                background-color: #f4f4f4;
                font-weight: bold;
            }
            .fancy-table tr:nth-child(even) {
                background-color: #f9f9f9;
            }
            .fancy-table caption {
                font-size: 1.5em;
                margin-bottom: 10px;
                font-weight: bold;
            }
        </style>
    </head>
    <body>

    <table>
        <thead>
            <tr>
                {% for key in data.keys() %}
                    <th><strong>{{ key }}</strong></th>
                {% endfor %}
            </tr>
        </thead>
        <tbody>
            <tr>
                {% for value in data.values() %}
                    <td>{{ value }}</td>
                {% endfor %}
            </tr>
        </tbody>
    </table>


    </body>
    </html>
    """
    )

    return template.render(data=data)


def to_html_construction(data: Dict[str, Any]) -> str:
    template = Template(
        """
    <!DOCTYPE html>
    <html lang="en">
    <head>
        <meta charset="UTF-8">
        <meta name="viewport" content="width=device-width, initial-scale=1.0">
        <title>Layer Information Table</title>
        <style>
            body {
                font-family: Arial, sans-serif;
                margin: 20px;
            }
            table {
                width: 100%;
                border-collapse: collapse;
                margin-bottom: 20px;
                font-size: 1em;
            }
            th, td {
                padding: 12px;
                border: 1px solid #ddd;
                text-align: left;
            }
            th {
                background-color: #f4f4f4;
                font-weight: bold;
            }
            tr:nth-child(even) {
                background-color: #f9f9f9;
            }
            caption {
                font-size: 0.75em;
                margin-bottom: 10px;
                font-weight: bold;
                text-align: left;
            }
        </style>
    </head>
    <body>

    <table>
        <caption>Layers for {{ data.name }}</caption>
        <thead>
            <tr>
                <th>Name</th>
                <th>c</th>
                <th>epsLw</th>
                <th>epsSw</th>
                <th>k</th>
                <th>rho</th>
                <th>Thickness</th>
            </tr>
        </thead>
        <tbody>
        {% for layer in data.layers %}
            <tr>
                <td>{{ layer.name }}</td>
                <td>{{ layer.c }}</td>
                <td>{{ layer.epsLw }}</td>
                <td>{{ layer.epsSw }}</td>
                <td>{{ layer.k }}</td>
                <td>{{ layer.rho }}</td>
                <td>{{ layer.thickness }}</td>
            </tr>
        {% endfor %}
        </tbody>
    </table>

    </body>
    </html>
    """
    )

    # Render the template with the data
    return template.render(data=data)


def get_figures(element_name: str, documentation: "ModelDocumentation") -> list:  # type: ignore
    # TODO: this feels like a duplicate with docx!!!! check why
    if documentation.result is None:
        return []
    mat = Reader(documentation.result.path, documentation.result.type)
    elements = [node for node in documentation.elements if node.name == element_name]
    if elements:
        element = elements[0]
        return plot_element(mat, element, plot_plot_ly)


def get_description() -> Dict[str, Any]:
    from trano.elements import BaseParameter

    return {
        field.alias: (field.description or field.title)
        for cls in BaseParameter.__subclasses__()
        for field in cls.model_fields.values()
        if field.alias
    }


def _get_elements(
    elements: List["BaseElement"],
    element_type: Type[Union["Space", "System", "BaseSimpleWall", "BaseWall"]],
) -> List["BaseElement"]:
    return [element for element in elements if isinstance(element, element_type)]


def _dump_list_attributes(
    element: "BaseElement", attribute_name: str, include_mapping: object
) -> List[Dict[int, Any]]:
    datas = []
    for el in getattr(element, attribute_name):
        data = el.model_dump(
            by_alias=True,
            exclude_none=True,
            mode="json",
            include=include_mapping,
        )
        if el.parameters:
            data["parameters"] = el.parameters.model_dump(
                by_alias=True, exclude_none=True
            )
        datas.append(data)
    return datas



from typing import TYPE_CHECKING, Any, List, Optional, Tuple, Type

from pydantic import BaseModel, Field, field_validator, model_validator

from trano import elements
from trano.elements.types import ConnectionView, Flow, PartialConnection
from trano.exceptions import IncompatiblePortsError

if TYPE_CHECKING:
    from trano.elements import BaseElement

INCOMPATIBLE_PORTS = [sorted(["dataBus", "y"])]


class Connection(BaseModel):
    right: PartialConnection
    left: PartialConnection
    connection_view: ConnectionView = Field(default=ConnectionView())

    @model_validator(mode="after")
    def _connection_validator(self) -> "Connection":
        if (
            sorted(
                [
                    part.split(".")[-1]
                    for part in [self.right.equation, self.left.equation]
                ]
            )
            in INCOMPATIBLE_PORTS
        ):
            raise IncompatiblePortsError(
                f"Incompatible ports {self.right.equation} and {self.left.equation}."
            )
        return self

    @property
    def path(self) -> List[List[float] | Tuple[float, float]]:
        if self.left.position[0] < self.right.position[0]:
            mid_path = (self.right.position[0] - self.left.position[0]) / 2
            return [
                self.left.position,
                (self.left.position[0] + mid_path, self.left.position[1]),
                (self.right.position[0] - mid_path, self.right.position[1]),
                self.right.position,
            ]

        else:
            mid_path = (self.left.position[0] - self.right.position[0]) / 2
            return [
                self.left.position,
                (self.left.position[0] - mid_path, self.left.position[1]),
                (self.right.position[0] + mid_path, self.right.position[1]),
                self.right.position,
            ]


class Port(BaseModel):
    names: list[str]
    targets: Optional[List[Any]] = None
    available: bool = True
    flow: Flow = Field(default=Flow.undirected)
    multi_connection: bool = False
    multi_object: bool = False
    bus_connection: bool = False
    use_counter: bool = True
    counter: int = Field(default=1)

    @field_validator("targets")
    @classmethod
    def validate_targets(cls, values: List[str]) -> List[Type["BaseElement"]]:
        from trano.elements import BaseElement

        targets: List[Type[BaseElement]] = []
        for value in values:
            if isinstance(value, str):
                if hasattr(elements, value):
                    targets.append(getattr(elements, value))
                else:
                    raise ValueError(f"Target {value} not found")
            else:
                targets.append(value)
        return targets

    def is_available(self) -> bool:
        return self.multi_connection or self.available

    def is_controllable(self) -> bool:
        from trano.elements.base import Control

        return self.targets is not None and any(
            target == Control for target in self.targets
        )

    def link(
        self, node: "BaseElement", connected_node: "BaseElement"
    ) -> list[PartialConnection]:

        from trano.elements.envelope import MergedBaseWall

        self.available = False
        partial_connections = []
        merged_number = 1
        if isinstance(node, MergedBaseWall):
            merged_number = len(node.surfaces)

        if isinstance(connected_node, MergedBaseWall):
            merged_number = len(connected_node.surfaces)
        for name in self.names:
            if self.multi_connection and self.bus_connection:
                first_counter = self.counter
                last_counter = self.counter + merged_number - 1
                counter = (
                    f"{first_counter}"
                    if first_counter == last_counter
                    else f"{first_counter}:{last_counter}"
                )
                equation = (
                    f"{node.name}[{counter}].{name}"
                    if self.multi_object
                    else f"{node.name}.{name}[{counter}]"
                )
                self.counter = last_counter + 1
            elif self.multi_connection and self.use_counter:
                equation = f"{node.name}.{name}[{self.counter}]"
                self.counter += 1
            else:
                equation = f"{node.name}.{name}"

            partial_connections.append(
                PartialConnection(equation=equation, position=node.position)
            )

        return partial_connections


def connection_color(edge: Tuple["BaseElement", "BaseElement"]) -> ConnectionView:
    from trano.elements.bus import DataBus
    from trano.elements.envelope import BaseSimpleWall
    from trano.elements.system import Weather

    if any(isinstance(e, BaseSimpleWall) for e in edge):
        return ConnectionView(color="{191,0,0}", thickness=0.1)
    if any(isinstance(e, DataBus) for e in edge):
        return ConnectionView(color=None, thickness=0.05)
    if any(isinstance(e, Weather) for e in edge):
        return ConnectionView(color=None, thickness=0.05)
    return ConnectionView()


def connect(edge: Tuple["BaseElement", "BaseElement"]) -> list[Connection]:
    connections = []
    edge_first = edge[0]
    edge_second = edge[1]
    edge_first_ports_to_skip: List[Port] = []
    edge_second_ports_to_skip: List[Port] = []
    while True:
        current_port = edge_first._get_target_compatible_port(
            edge_second, Flow.outlet, ports_to_skip=edge_first_ports_to_skip
        )
        other_port = edge_second._get_target_compatible_port(
            edge_first, Flow.inlet, ports_to_skip=edge_second_ports_to_skip
        )
        if any(port is None for port in [current_port, other_port]):
            break
        for left, right in zip(
            current_port.link(edge_first, edge_second),  # type: ignore
            other_port.link(edge_second, edge_first),  # type: ignore
        ):
            connections.append(
                Connection(
                    left=left, right=right, connection_view=connection_color(edge)
                )
            )
        edge_first_ports_to_skip.append(current_port)  # type: ignore
        edge_second_ports_to_skip.append(other_port)  # type: ignore
    return connections


def _has_inlet_or_outlet(target: "BaseElement") -> bool:
    return bool([port for port in target.ports if port.flow == Flow.inlet_or_outlet])


def _is_inlet_or_outlet(target: "BaseElement") -> bool:
    return bool(
        [port for port in target.ports if port.flow in [Flow.inlet, Flow.outlet]]
    )



from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Optional, Union

from jinja2 import Environment, FileSystemLoader
from networkx.classes.reportviews import NodeView
from pydantic import BaseModel, ConfigDict, Field, field_validator

if TYPE_CHECKING:

    from trano.library.library import Library


class Material(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    name: str
    thermal_conductivity: float = Field(
        ..., title="Thermal conductivity [W/(m.K)]", alias="k"
    )
    specific_heat_capacity: float = Field(
        ..., title="Specific thermal capacity [J/(kg.K)]", alias="c"
    )
    density: float = Field(..., title="Density [kg/m3]", alias="rho")
    longwave_emissivity: float = Field(
        0.85, title="Longwave emissivity [1]", alias="epsLw"
    )
    shortwave_emissivity: float = Field(
        0.65, title="Shortwave emissivity [1]", alias="epsSw"
    )

    def __hash__(self) -> int:
        return hash(self.name)

    @field_validator("name")
    @classmethod
    def clean_name(cls, value: str) -> str:
        if ":" in value:
            return value.lower().replace(":", "_")
        return value


class GlassMaterial(Material):
    solar_transmittance: list[float]
    solar_reflectance_outside_facing: list[float]
    solar_reflectance_room_facing: list[float]
    infrared_transmissivity: float
    infrared_absorptivity_outside_facing: float
    infrared_absorptivity_room_facing: float


class Gas(Material): ...


class Layer(BaseModel):
    material: Material
    thickness: float


class Construction(BaseModel):
    name: str
    layers: list[Layer]

    def __hash__(self) -> int:
        return hash(self.name)

    @field_validator("name")
    @classmethod
    def clean_name(cls, value: str) -> str:
        if ":" in value:
            return value.lower().replace(":", "_")
        return value


class GlassLayer(BaseModel):
    thickness: float
    material: GlassMaterial
    layer_type: str = "glass"


class GasLayer(BaseModel):
    model_config = ConfigDict(use_enum_values=True)
    thickness: float
    material: Gas
    layer_type: str = "gas"


class Glass(BaseModel):
    name: str
    layers: list[GlassLayer | GasLayer]
    u_value_frame: float

    def __hash__(self) -> int:
        return hash(self.name)

    @field_validator("name")
    @classmethod
    def clean_name(cls, value: str) -> str:
        if ":" in value:
            return value.lower().replace(":", "_")
        return value


class BaseData(BaseModel):
    template: Optional[str] = None
    constructions: List[Union[Construction, Material, Glass]]


class ConstructionData(BaseModel):
    constructions: List[Construction]
    materials: List[Material]
    glazing: List[Glass]


class BaseConstructionData(BaseModel):
    template: str
    construction: BaseData
    material: BaseData
    glazing: BaseData

    def generate_data(self, package_name: str) -> str:
        environment = Environment(
            trim_blocks=True,
            lstrip_blocks=True,
            loader=FileSystemLoader(
                str(Path(__file__).parents[1].joinpath("templates"))
            ),
            autoescape=True,
        )
        models: Dict[str, List[str]] = {
            "material": [],
            "construction": [],
            "glazing": [],
        }
        for construction_type_name in models:
            construction_type = getattr(self, construction_type_name)
            for construction in construction_type.constructions:
                template = environment.from_string(
                    "{% import 'macros.jinja2' as macros %}"
                    + construction_type.template
                )
                model = template.render(
                    construction=construction, package_name=package_name
                )
                models[construction_type_name].append(model)
        template = environment.from_string(
            "{% import 'macros.jinja2' as macros %}" + self.template
        )
        model = template.render(**models, package_name=package_name)
        return model


class MaterialProperties(BaseModel):
    data: str
    is_package: bool


class BaseTemplateData(BaseModel):
    template: Optional[str] = None
    constructions: List[Union[Construction, Material, Glass]]


def default_construction(nodes: NodeView) -> ConstructionData:
    from trano.elements.envelope import BaseSimpleWall

    constructions = {
        node.construction
        for node in [node_ for node_ in nodes if isinstance(node_, BaseSimpleWall)]
    }
    wall_constructions = sorted(
        [c for c in constructions if isinstance(c, Construction)], key=lambda x: x.name
    )
    glazing = sorted(
        [c for c in constructions if isinstance(c, Glass)], key=lambda x: x.name
    )
    return ConstructionData(
        constructions=wall_constructions, materials=[], glazing=glazing
    )


def merged_construction(nodes: NodeView) -> ConstructionData:
    # TODO: Fix the import
    from trano.elements.envelope import BaseSimpleWall, MergedBaseWall

    merged_constructions = {
        construction
        for node in [node_ for node_ in nodes if isinstance(node_, MergedBaseWall)]
        for construction in node.constructions
    }
    constructions = {
        node.construction
        for node in [node_ for node_ in nodes if isinstance(node_, BaseSimpleWall)]
    }
    merged_constructions.update(constructions)
    wall_constructions = [
        c for c in merged_constructions if isinstance(c, Construction)
    ]
    glazing = [c for c in merged_constructions if isinstance(c, Glass)]
    materials = {
        layer.material
        for construction in merged_constructions
        for layer in construction.layers
    }
    return ConstructionData(
        constructions=wall_constructions, materials=list(materials), glazing=glazing
    )


def extract_data(
    package_name: str, nodes: NodeView, library: "Library"
) -> MaterialProperties:

    data = (
        merged_construction(nodes)
        if library.merged_external_boundaries
        else default_construction(nodes)
    )
    data_ = BaseConstructionData(
        template=library.templates.main,
        construction=BaseData(
            constructions=data.constructions, template=library.templates.construction
        ),
        glazing=BaseData(
            constructions=data.glazing, template=library.templates.glazing
        ),
        material=BaseData(
            constructions=data.materials, template=library.templates.material
        ),
    )
    return MaterialProperties(
        data=data_.generate_data(package_name), is_package=library.templates.is_package
    )


def extract_properties(
    library: "Library", package_name: str, nodes: NodeView
) -> MaterialProperties:
    return extract_data(package_name, nodes, library)



from trano.elements.base import BaseElement, Control
from trano.elements.boundary import BaseBoundary, Boundary
from trano.elements.bus import DataBus
from trano.elements.connection import Connection, Port, connect
from trano.elements.control import (
    AhuControl,
    BoilerControl,
    CollectorControl,
    EmissionControl,
    ThreeWayValveControl,
    VAVControl,
)
from trano.elements.envelope import (
    BaseInternalElement,
    BaseSimpleWall,
    BaseWall,
    ExternalDoor,
    ExternalWall,
    FloorOnGround,
    InternalElement,
    WallParameters,
    Window,
)
from trano.elements.figure import Figure, NamedFigure
from trano.elements.parameters import BaseParameter, param_from_config
from trano.elements.space import Space
from trano.elements.system import (
    VAV,
    AirHandlingUnit,
    BaseOccupancy,
    BaseWeather,
    Boiler,
    Duct,
    Emission,
    Pump,
    Radiator,
    SplitValve,
    System,
    TemperatureSensor,
    ThreeWayValve,
    Valve,
    Ventilation,
)
from trano.elements.types import DynamicTemplateCategories

__all__ = [
    "ThreeWayValveControl",
    "SplitValve",
    "CollectorControl",
    "BaseInternalElement",
    "BoilerControl",
    "BaseOccupancy",
    "Emission",
    "VAVControl",
    "BaseWall",
    "ThreeWayValve",
    "TemperatureSensor",
    "Boundary",
    "DataBus",
    "Space",
    "BaseBoundary",
    "System",
    "AhuControl",
    "BaseWeather",
    "AirHandlingUnit",
    "Ventilation",
    "param_from_config",
    "ExternalDoor",
    "ExternalWall",
    "FloorOnGround",
    "InternalElement",
    "Window",
    "BaseElement",
    "BaseSimpleWall",
    "WallParameters",
    "DynamicTemplateCategories",
    "Figure",
    "Control",
    "connect",
    "Connection",
    "BaseParameter",
    "Port",
    "VAV",
    "Duct",
    "Radiator",
    "Valve",
    "EmissionControl",
    "Boiler",
    "Pump",
    "NamedFigure",
]



from typing import List, Optional

from trano.elements import Control
from trano.elements.base import BaseElement
from trano.elements.types import BaseVariant


class System(BaseElement):
    position: Optional[List[float]] = None
    control: Optional[Control] = None


class Sensor(System): ...


class EmissionVariant(BaseVariant):
    radiator: str = "radiator"
    ideal: str = "ideal"


class SpaceSystem(System):
    linked_space: Optional[str] = None


class Emission(SpaceSystem): ...


class Ventilation(SpaceSystem): ...


class BaseWeather(System): ...


class BaseOccupancy(System):
    space_name: Optional[str] = None


class Weather(BaseWeather): ...


class Valve(SpaceSystem): ...


class ThreeWayValve(System): ...


class TemperatureSensor(Sensor): ...


class SplitValve(System): ...


class Radiator(Emission): ...


class Pump(System): ...


class Occupancy(BaseOccupancy): ...


class Duct(Ventilation): ...


class DamperVariant(BaseVariant):
    complex: str = "complex"


class Damper(Ventilation): ...


class VAV(Damper):
    variant: str = DamperVariant.default


class Boiler(System): ...


class AirHandlingUnit(Ventilation): ...



import logging.config

LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "standard": {
            "format": "%(asctime)s [%(levelname)s] %(name)s: [%(funcName)s] %(message)s"
        },
    },
    "handlers": {
        "console": {
            "level": "ERROR",
            "class": "logging.StreamHandler",
            "formatter": "standard",
        },
        "file": {
            "level": "DEBUG",
            "class": "logging.FileHandler",
            "formatter": "standard",
            "filename": "trano.log",
            "mode": "w+",
        },
    },
    "loggers": {
        "": {
            "handlers": ["console", "file"],
            "level": "DEBUG",
            "propagate": True,
        },
    },
}


logging.config.dictConfig(LOGGING_CONFIG)
logging.getLogger("docker").setLevel(logging.CRITICAL)
logging.getLogger("urllib3").setLevel(logging.CRITICAL)



class WrongSystemFlowError(Exception):
    """Exception raised for wrong system flow."""


class ControllerBusPortError(Exception):
    """Exception raised for controller bus port error."""


class IncompatiblePortsError(Exception): ...


class DockerNotInstalledError(Exception): ...


class DockerClientError(Exception): ...



import json
import os
from pathlib import Path
from typing import IO, Any

import yaml


class Loader(yaml.SafeLoader):
    """YAML Loader with `!include` constructor."""

    def __init__(self, stream: IO) -> None:  # type: ignore
        """Initialise Loader."""

        try:
            self._root = os.path.split(stream.name)[0]
        except AttributeError:
            self._root = os.path.curdir

        super().__init__(stream)


def include_default(loader: Loader, node: yaml.Node) -> Any:  # noqa: ANN401
    filename = Path(__file__).parent.joinpath("default.yaml")

    return _load(filename)


def _load(filename: Path) -> Any:  # noqa: ANN401
    with filename.open("r") as f:
        if filename.suffix in (".yaml", ".yml"):
            return yaml.load(f, Loader)  # noqa: S506
        elif filename.suffix == ".json":
            return json.load(f)
        else:
            return "".join(f.readlines())


yaml.add_constructor(
    "!include_default",
    include_default,
    Loader,
)



from typing import List, Optional

from pydantic import BaseModel, computed_field, field_validator


class Target(BaseModel):
    main: str
    sub: Optional[str] = None
    evaluated_element: str = ""
    value: str = ""

    @field_validator("value")
    @classmethod
    def validate_value(cls, value: str) -> str:
        return value.capitalize()

    def __hash__(self) -> int:
        return hash((self.main, self.sub))

    def commands(self) -> list[str]:
        return self.main.split(".")[1:]

    def sub_commands(self) -> List[str]:
        if self.sub is None:
            raise Exception("Target sub is None")
        return self.sub.split(".")


class BaseInputOutput(BaseModel):
    name: str
    component: str
    port: str
    multi: bool = False
    target: Target
    input_template: str
    default: float | str | int

    def __hash__(self) -> int:
        return hash((self.name, self.target.value))

    def __eq__(self, other: object) -> bool:
        return hash(self) == hash(other)

    @computed_field  # type: ignore
    @property
    def input_model(self) -> str:
        if self.target.evaluated_element:
            return f"""{self.input_template}
            {self.name}{self.target.evaluated_element.capitalize()}
            (y={self.default});"""
        return ""


class BaseInput(BaseInputOutput): ...


class RealInput(BaseInput):
    default: float = 0.0
    input_template: str = "Modelica.Blocks.Sources.RealExpression"


class IntegerInput(BaseInput):
    default: int = 0
    input_template: str = "Modelica.Blocks.Sources.IntegerExpression"


class BooleanInput(BaseInput):
    default: str = "false"
    input_template: str = "Modelica.Blocks.Sources.BooleanExpression"


class BooleanOutput(BaseInputOutput):
    default: str = "false"
    input_template: str = "Modelica.Blocks.Sources.BooleanExpression"


class IntegerOutput(BaseInputOutput):
    default: int = 0
    input_template: str = "Modelica.Blocks.Sources.IntegerExpression"


class RealOutput(BaseInputOutput):
    default: float = 0.0
    input_template: str = "Modelica.Blocks.Sources.RealExpression"



import tempfile
import webbrowser
from enum import Enum
from pathlib import Path
from typing import Optional, Annotated
from rich import print
import typer
from trano.data_models.conversion import convert_network
from trano.library.library import Library
from trano.reporting.html import to_html_reporting
from trano.reporting.reporting import ModelDocumentation
from trano.reporting.types import ResultFile
from trano.simulate.simulate import SimulationLibraryOptions, simulate
from trano.topology import Network
from trano.utils.utils import is_success
from rich.progress import Progress, SpinnerColumn, TextColumn

app = typer.Typer()
CHECKMARK = "[green]✔[/green]"
CROSS_MARK = "[red]✘[/red]"


class LibraryChoice(str, Enum):
    ideas = "IDEAS"
    buildings = "Buildings"


def _create_network(model: str, library: str) -> Network:
    library_ = Library.from_configuration(library)
    model_ = Path(model).resolve()
    return convert_network(str(model_.stem), model_, library=library_)


@app.command()
def create_model(
    model: Annotated[
        str,
        typer.Argument(help="Local path to the '.yaml' model configuration file"),
    ],
    library: Annotated[
        LibraryChoice,
        typer.Argument(help="Library to be used for simulation."),
    ] = LibraryChoice.buildings,
) -> None:
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        transient=True,
    ) as progress:
        modelica_model_path = Path(model).resolve().with_suffix(".mo")
        task = progress.add_task(
            description=f"Generating model {modelica_model_path.name} with library {library}",
            total=None,
        )
        network = _create_network(model, library)
        modelica_model = network.model()
        progress.update(task, completed=True)
        task = progress.add_task(description="Writing model to file...", total=None)
        modelica_model_path.write_text(modelica_model)
        progress.remove_task(task)
        print(f"{CHECKMARK} Model generated at {modelica_model_path}")


@app.command()
def simulate_model(
    model: Annotated[
        str,
        typer.Argument(help="Local path to the '.yaml' model configuration file."),
    ],
    library: Annotated[
        Optional[LibraryChoice],
        typer.Argument(help="Library to be used for simulation."),
    ] = LibraryChoice.buildings,
    start: Annotated[Optional[int], typer.Argument(help="Start simulation time.")] = 0,
    end: Annotated[Optional[int], typer.Argument(help="End simulation time.")] = 2
    * 3600
    * 24
    * 7,
    tolerance: Annotated[
        Optional[float], typer.Argument(help="Simulation tolerance.")
    ] = 1e-4,
) -> None:
    options = SimulationLibraryOptions(
        start_time=start,
        end_time=end,
        tolerance=tolerance,
        library_name=library,
    )
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        transient=True,
    ) as progress:
        task = progress.add_task(
            description=f"Generating model {Path(model).stem} with library {options.library_name}",
            total=None,
        )
        network = _create_network(model, options.library_name)
        model_ = Path(model).resolve()
        progress.remove_task(task)
        print(f"{CHECKMARK} Model generated successfully.")
        task = progress.add_task(
            description="Simulating model ...",
            total=None,
        )
        results = simulate(
            model_.parent,
            network,
            options=options,
        )

        if not is_success(results):
            print(f"{CROSS_MARK} Simulation failed. See logs for more information.")
            return

        result_path = (
            Path(model_.parent) / "results" / f"{model_.stem}.building_res.mat"
        )
        if not result_path.exists():
            print(
                f"{CROSS_MARK} Simulation failed. Result file not found in {result_path}."
            )
            return
        progress.remove_task(task)
        print(f"{CHECKMARK} Simulation results available at {result_path}")

        task = progress.add_task(
            description="Creating report ...",
            total=None,
        )
        reporting = ModelDocumentation.from_network(
            network,
            result=ResultFile(path=result_path),
        )
        html = to_html_reporting(reporting)
        report_path = Path(model_.parent / f"{model_.stem}.html")
        report_path.write_text(html)
        webbrowser.open(f"file://{report_path}")
        progress.remove_task(task)
        print(f"{CHECKMARK} Report available at {report_path}")


@app.command()
def verify() -> None:
    model_path = Path(__file__).parent / "verification.yaml"
    options = SimulationLibraryOptions(end_time=3600)
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        transient=True,
    ) as progress:
        task = progress.add_task(
            description="Verify generating model...",
            total=None,
        )
        network = _create_network(str(model_path), options.library_name)

        progress.remove_task(
            task,
        )
        print(
            f"{CHECKMARK} Model generated successfully. Your system is compatible for model generation."
        )
        task_ = progress.add_task(
            description="Verify simulation...",
            total=None,
        )
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as project_path:
            try:
                simulate(
                    Path(project_path),
                    network,
                    options=options,
                )
            except Exception as e:
                print(f"{CROSS_MARK} Simulation failed. Reason {e}.")
                return
        progress.remove_task(
            task_,
        )
        print(
            f"{CHECKMARK} Model simulated successfully. Your system is compatible for simulation."
        )


def report(model: Path | str, options: SimulationLibraryOptions) -> None:
    model = Path(model).resolve()
    network = convert_network(
        model.stem, model, library=Library.from_configuration(options.library_name)
    )
    reporting = ModelDocumentation.from_network(
        network,
        result=ResultFile(
            path=Path(model.parent) / "results" / f"{model.stem}.building_res.mat"
        ),
    )
    html = to_html_reporting(reporting)
    report_path = Path(model.parent / f"{model.stem}.html")
    report_path.write_text(html)
    webbrowser.open(f"file://{report_path}")


if __name__ == "__main__":
    app()
